@@ -2,7 +2,7 @@
  * ccache -- a fast C/C++ compiler cache
  *
  * Copyright (C) 2002-2007 Andrew Tridgell
- * Copyright (C) 2009-2010 Joel Rosdahl
+ * Copyright (C) 2009-2012 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -36,7 +36,7 @@ static const char VERSION_TEXT[] =
 MYNAME " version %s\n"
 "\n"
 "Copyright (C) 2002-2007 Andrew Tridgell\n"
-"Copyright (C) 2009-2010 Joel Rosdahl\n"
+"Copyright (C) 2009-2011 Joel Rosdahl\n"
 "\n"
 "This program is free software; you can redistribute it and/or modify it under\n"
 "the terms of the GNU General Public License as published by the Free Software\n"
@@ -67,7 +67,7 @@ static const char USAGE_TEXT[] =
 "See also <http://ccache.samba.org>.\n";
 
 /* current working directory taken from $PWD, or getcwd() if $PWD is bad */
-static char *current_working_dir;
+char *current_working_dir = NULL;
 
 /* the base cache directory */
 char *cache_dir = NULL;
@@ -79,7 +79,7 @@ static char *temp_dir;
 char *cache_logfile = NULL;
 
 /* base directory (from CCACHE_BASEDIR) */
-static char *base_dir;
+char *base_dir = NULL;
 
 /* the original argument list */
 static struct args *orig_args;
@@ -267,8 +267,7 @@ get_path_in_cache(const char *name, const char *suffix)
 		free(path);
 		path = p;
 		if (create_dir(path) != 0) {
-			cc_log("Failed to create %s", path);
-			failed();
+			fatal("Failed to create %s: %s", path, strerror(errno));
 		}
 	}
 	result = format("%s/%s%s", path, name + nlevels, suffix);
@@ -307,7 +306,7 @@ remember_include_file(char *path, size_t path_len, struct mdfour *cpp_hash)
 	}
 
 	if (stat(path, &st) != 0) {
-		cc_log("Failed to stat include file %s", path);
+		cc_log("Failed to stat include file %s: %s", path, strerror(errno));
 		goto failure;
 	}
 	if (S_ISDIR(st.st_mode)) {
@@ -381,18 +380,39 @@ ignore:
 }
 
 /*
- * Make a relative path from CCACHE_BASEDIR to path. Takes over ownership of
- * path. Caller frees.
+ * Make a relative path from current working directory to path if CCACHE_PREFIX
+ * is a prefix of path. Takes over ownership of path. Caller frees.
  */
 static char *
 make_relative_path(char *path)
 {
-	char *relpath;
+	char *relpath, *canon_path;
 
 	if (!base_dir || !str_startswith(path, base_dir)) {
 		return path;
 	}
 
+	if (!current_working_dir) {
+		char *cwd = get_cwd();
+		if (cwd) {
+			current_working_dir = x_realpath(cwd);
+			free(cwd);
+		}
+		if (!current_working_dir) {
+			cc_log("Unable to determine current working directory: %s",
+			       strerror(errno));
+			failed();
+		}
+	}
+
+	canon_path = x_realpath(path);
+	if (canon_path) {
+		free(path);
+		path = canon_path;
+	} else {
+		/* path doesn't exist, so leave it as it is. */
+	}
+
 	relpath = get_relative_path(current_working_dir, path);
 	free(path);
 	return relpath;
@@ -530,7 +550,10 @@ to_cache(struct args *args)
 	status = execute(args->argv, tmp_stdout, tmp_stderr);
 	args_pop(args, 3);
 
-	if (stat(tmp_stdout, &st) != 0 || st.st_size != 0) {
+	if (stat(tmp_stdout, &st) != 0) {
+		fatal("Could not create %s (permission denied?)", tmp_stdout);
+	}
+	if (st.st_size != 0) {
 		cc_log("Compiler produced stdout");
 		stats_update(STATS_STDOUT);
 		tmp_unlink(tmp_stdout);
@@ -552,22 +575,23 @@ to_cache(struct args *args)
 
 		tmp_stderr2 = format("%s.tmp.stderr2.%s", cached_obj, tmp_string());
 		if (x_rename(tmp_stderr, tmp_stderr2)) {
-			cc_log("Failed to rename %s to %s", tmp_stderr, tmp_stderr2);
+			cc_log("Failed to rename %s to %s: %s", tmp_stderr, tmp_stderr2,
+			       strerror(errno));
 			failed();
 		}
 		fd_cpp_stderr = open(cpp_stderr, O_RDONLY | O_BINARY);
 		if (fd_cpp_stderr == -1) {
-			cc_log("Failed opening %s", cpp_stderr);
+			cc_log("Failed opening %s: %s", cpp_stderr, strerror(errno));
 			failed();
 		}
 		fd_real_stderr = open(tmp_stderr2, O_RDONLY | O_BINARY);
 		if (fd_real_stderr == -1) {
-			cc_log("Failed opening %s", tmp_stderr2);
+			cc_log("Failed opening %s: %s", tmp_stderr2, strerror(errno));
 			failed();
 		}
 		fd_result = open(tmp_stderr, O_WRONLY | O_CREAT | O_TRUNC | O_BINARY, 0666);
 		if (fd_result == -1) {
-			cc_log("Failed opening %s", tmp_stderr);
+			cc_log("Failed opening %s: %s", tmp_stderr, strerror(errno));
 			failed();
 		}
 		copy_fd(fd_cpp_stderr, fd_result);
@@ -615,14 +639,15 @@ to_cache(struct args *args)
 	}
 
 	if (stat(tmp_stderr, &st) != 0) {
-		cc_log("Failed to stat %s", tmp_stderr);
+		cc_log("Failed to stat %s: %s", tmp_stderr, strerror(errno));
 		stats_update(STATS_ERROR);
 		failed();
 	}
 	if (st.st_size > 0) {
 		if (move_uncompressed_file(tmp_stderr, cached_stderr,
 		                           enable_compression) != 0) {
-			cc_log("Failed to move %s to %s", tmp_stderr, cached_stderr);
+			cc_log("Failed to move %s to %s: %s", tmp_stderr, cached_stderr,
+			       strerror(errno));
 			stats_update(STATS_ERROR);
 			failed();
 		}
@@ -636,7 +661,7 @@ to_cache(struct args *args)
 		tmp_unlink(tmp_stderr);
 	}
 	if (move_uncompressed_file(tmp_obj, cached_obj, enable_compression) != 0) {
-		cc_log("Failed to move %s to %s", tmp_obj, cached_obj);
+		cc_log("Failed to move %s to %s: %s", tmp_obj, cached_obj, strerror(errno));
 		stats_update(STATS_ERROR);
 		failed();
 	} else {
@@ -651,7 +676,7 @@ to_cache(struct args *args)
 	 * size statistics.
 	 */
 	if (stat(cached_obj, &st) != 0) {
-		cc_log("Failed to stat %s", strerror(errno));
+		cc_log("Failed to stat %s: %s", cached_obj, strerror(errno));
 		stats_update(STATS_ERROR);
 		failed();
 	}
@@ -708,8 +733,8 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 		   correct i_tmpfile */
 		path_stdout = input_file;
 		if (create_empty_file(path_stderr) != 0) {
+			cc_log("Failed to create %s: %s", path_stderr, strerror(errno));
 			stats_update(STATS_ERROR);
-			cc_log("Failed to create %s", path_stderr);
 			failed();
 		}
 		status = 0;
@@ -751,7 +776,7 @@ get_object_name_from_cpp(struct args *args, struct mdfour *hash)
 
 	hash_delimiter(hash, "cppstderr");
 	if (!hash_file(hash, path_stderr)) {
-		fatal("Failed to open %s", path_stderr);
+		fatal("Failed to open %s: %s", path_stderr, strerror(errno));
 	}
 
 	i_tmpfile = path_stdout;
@@ -788,6 +813,36 @@ update_cached_result_globals(struct file_hash *hash)
 	free(object_name);
 }
 
+/*
+ * Hash mtime or content of a file, or the output of a command, according to
+ * the CCACHE_COMPILERCHECK setting.
+ */
+static void
+hash_compiler(struct mdfour *hash, struct stat *st, const char *path,
+              bool allow_command)
+{
+	const char *compilercheck;
+
+	compilercheck = getenv("CCACHE_COMPILERCHECK");
+	if (!compilercheck) {
+		compilercheck = "mtime";
+	}
+	if (str_eq(compilercheck, "none")) {
+		/* Do nothing. */
+	} else if (str_eq(compilercheck, "mtime")) {
+		hash_delimiter(hash, "cc_mtime");
+		hash_int(hash, st->st_size);
+		hash_int(hash, st->st_mtime);
+	} else if (str_eq(compilercheck, "content") || !allow_command) {
+		hash_delimiter(hash, "cc_content");
+		hash_file(hash, path);
+	} else { /* command string */
+		if (!hash_multicommand_output(hash, compilercheck, orig_args->argv[0])) {
+			fatal("Failure running compiler check command: %s", compilercheck);
+		}
+	}
+}
+
 /*
  * Update a hash sum with information common for the direct and preprocessor
  * modes.
@@ -796,7 +851,6 @@ static void
 calculate_common_hash(struct args *args, struct mdfour *hash)
 {
 	struct stat st;
-	const char *compilercheck;
 	char *p;
 
 	hash_string(hash, HASH_PREFIX);
@@ -809,7 +863,7 @@ calculate_common_hash(struct args *args, struct mdfour *hash)
 	hash_string(hash, i_extension);
 
 	if (stat(args->argv[0], &st) != 0) {
-		cc_log("Couldn't stat the compiler (%s)", args->argv[0]);
+		cc_log("Couldn't stat compiler %s: %s", args->argv[0], strerror(errno));
 		stats_update(STATS_COMPILER);
 		failed();
 	}
@@ -817,31 +871,16 @@ calculate_common_hash(struct args *args, struct mdfour *hash)
 	/*
 	 * Hash information about the compiler.
 	 */
-	compilercheck = getenv("CCACHE_COMPILERCHECK");
-	if (!compilercheck) {
-		compilercheck = "mtime";
-	}
-	if (str_eq(compilercheck, "none")) {
-		/* Do nothing. */
-	} else if (str_eq(compilercheck, "content")) {
-		hash_delimiter(hash, "cc_content");
-		hash_file(hash, args->argv[0]);
-	} else if (str_eq(compilercheck, "mtime")) {
-		hash_delimiter(hash, "cc_mtime");
-		hash_int(hash, st.st_size);
-		hash_int(hash, st.st_mtime);
-	} else { /* command string */
-		if (!hash_multicommand_output(hash, compilercheck, orig_args->argv[0])) {
-			fatal("Failure running compiler check command: %s", compilercheck);
-		}
-	}
+	hash_compiler(hash, &st, args->argv[0], true);
 
 	/*
 	 * Also hash the compiler name as some compilers use hard links and
 	 * behave differently depending on the real name.
 	 */
 	hash_delimiter(hash, "cc_name");
-	hash_string(hash, basename(args->argv[0]));
+	p = basename(args->argv[0]);
+	hash_string(hash, p);
+	free(p);
 
 	/* Possibly hash the current working directory. */
 	if (getenv("CCACHE_HASHDIR")) {
@@ -884,6 +923,7 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 	struct stat st;
 	int result;
 	struct file_hash *object_hash = NULL;
+	char *p;
 
 	/* first the arguments */
 	for (i = 1; i < args->argc; i++) {
@@ -911,14 +951,34 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 			}
 		}
 
-		if (str_startswith(args->argv[i], "--specs=") &&
-		    stat(args->argv[i] + 8, &st) == 0) {
-			/* If given a explicit specs file, then hash that file,
+		p = NULL;
+		if (str_startswith(args->argv[i], "-specs=")) {
+			p = args->argv[i] + 7;
+		} else if (str_startswith(args->argv[i], "--specs=")) {
+			p = args->argv[i] + 8;
+		}
+		if (p && stat(p, &st) == 0) {
+			/* If given an explicit specs file, then hash that file,
 			   but don't include the path to it in the hash. */
 			hash_delimiter(hash, "specs");
-			if (!hash_file(hash, args->argv[i] + 8)) {
-				failed();
-			}
+			hash_compiler(hash, &st, p, false);
+			continue;
+		}
+
+		if (str_startswith(args->argv[i], "-fplugin=")
+		    && stat(args->argv[i] + 9, &st) == 0) {
+			hash_delimiter(hash, "plugin");
+			hash_compiler(hash, &st, args->argv[i] + 9, false);
+			continue;
+		}
+
+		if (str_eq(args->argv[i], "-Xclang")
+		    && i + 3 < args->argc
+		    && str_eq(args->argv[i+1], "-load")
+		    && str_eq(args->argv[i+2], "-Xclang")
+		    && stat(args->argv[i+3], &st) == 0) {
+			hash_delimiter(hash, "plugin");
+			hash_compiler(hash, &st, args->argv[i+3], false);
 			continue;
 		}
 
@@ -928,6 +988,24 @@ calculate_object_hash(struct args *args, struct mdfour *hash, int direct_mode)
 	}
 
 	if (direct_mode) {
+		/* Hash environment variables that affect the preprocessor output. */
+		const char **p;
+		const char *envvars[] = {
+			"CPATH",
+			"C_INCLUDE_PATH",
+			"CPLUS_INCLUDE_PATH",
+			"OBJC_INCLUDE_PATH",
+			"OBJCPLUS_INCLUDE_PATH", /* clang */
+			NULL
+		};
+		for (p = envvars; *p != NULL ; ++p) {
+			char *v = getenv(*p);
+			if (v) {
+				hash_delimiter(hash, *p);
+				hash_string(hash, v);
+			}
+		}
+
 		if (!(sloppiness & SLOPPY_FILE_MACRO)) {
 			/*
 			 * The source code file or an include file may contain
@@ -1022,7 +1100,7 @@ from_cache(enum fromcache_call_mode mode, bool put_object_in_manifest)
 			cc_log("Object file %s just disappeared from cache", cached_obj);
 			stats_update(STATS_MISSING);
 		} else {
-			cc_log("Failed to copy/link %s to %s (%s)",
+			cc_log("Failed to copy/link %s to %s: %s",
 			       cached_obj, output_obj, strerror(errno));
 			stats_update(STATS_ERROR);
 			failed();
@@ -1053,9 +1131,8 @@ from_cache(enum fromcache_call_mode mode, bool put_object_in_manifest)
 				cc_log("Dependency file %s just disappeared from cache", output_obj);
 				stats_update(STATS_MISSING);
 			} else {
-				cc_log("Failed to copy/link %s to %s (%s)",
-				       cached_dep, output_dep,
-				       strerror(errno));
+				cc_log("Failed to copy/link %s to %s: %s",
+				       cached_dep, output_dep, strerror(errno));
 				stats_update(STATS_ERROR);
 				failed();
 			}
@@ -1082,7 +1159,8 @@ from_cache(enum fromcache_call_mode mode, bool put_object_in_manifest)
 		/* Store the dependency file in the cache. */
 		ret = copy_file(output_dep, cached_dep, enable_compression);
 		if (ret == -1) {
-			cc_log("Failed to copy %s to %s", output_dep, cached_dep);
+			cc_log("Failed to copy %s to %s: %s", output_dep, cached_dep,
+			       strerror(errno));
 			/* Continue despite the error. */
 		} else {
 			cc_log("Stored in cache: %s", cached_dep);
@@ -1237,6 +1315,13 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 			continue;
 		}
 
+		/* Special case for -E. */
+		if (str_eq(argv[i], "-E")) {
+			stats_update(STATS_PREPROCESSING);
+			result = false;
+			goto out;
+		}
+
 		/* These are always too hard. */
 		if (compopt_too_hard(argv[i])
 		    || str_startswith(argv[i], "@")
@@ -1317,14 +1402,14 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 				result = false;
 				goto out;
 			}
-			output_obj = argv[i+1];
+			output_obj = make_relative_path(x_strdup(argv[i+1]));
 			i++;
 			continue;
 		}
 
 		/* alternate form of -o, with no space */
 		if (str_startswith(argv[i], "-o")) {
-			output_obj = &argv[i][2];
+			output_obj = make_relative_path(x_strdup(&argv[i][2]));
 			continue;
 		}
 
@@ -1356,34 +1441,73 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 			args_add(dep_args, argv[i]);
 			continue;
 		}
-		if (i < argc - 1) {
-			if (str_startswith(argv[i], "-MF")) {
-				char *arg;
-				dependency_filename_specified = true;
-				free(output_dep);
-				args_add(dep_args, argv[i]);
-				if (strlen(argv[i]) == 3) {
-					/* -MF arg */
-					arg = argv[i + 1];
-					args_add(dep_args, argv[i + 1]);
-					i++;
-				} else {
-					/* -MFarg */
-					arg = &argv[i][3];
+		if (str_startswith(argv[i], "-MF")) {
+			char *arg;
+			bool separate_argument = (strlen(argv[i]) == 3);
+			dependency_filename_specified = true;
+			free(output_dep);
+			if (separate_argument) {
+				/* -MF arg */
+				if (i >= argc - 1) {
+					cc_log("Missing argument to %s", argv[i]);
+					stats_update(STATS_ARGS);
+					result = false;
+					goto out;
 				}
-				output_dep = make_relative_path(x_strdup(arg));
-				continue;
-			} else if (str_startswith(argv[i], "-MQ")
-			           || str_startswith(argv[i], "-MT")) {
-				dependency_target_specified = true;
-				args_add(dep_args, argv[i]);
-				if (strlen(argv[i]) == 3) {
-					/* -MQ arg or -MT arg */
-					args_add(dep_args, argv[i + 1]);
-					i++;
-				}
-				continue;
+				arg = argv[i + 1];
+				i++;
+			} else {
+				/* -MFarg */
+				arg = &argv[i][3];
 			}
+			output_dep = make_relative_path(x_strdup(arg));
+			/* Keep the format of the args the same */
+			if (separate_argument) {
+				args_add(dep_args, "-MF");
+				args_add(dep_args, output_dep);
+			} else {
+				char *option = format("-MF%s", output_dep);
+				args_add(dep_args, option);
+				free(option);
+			}
+			continue;
+		}
+		if (str_startswith(argv[i], "-MQ") || str_startswith(argv[i], "-MT")) {
+			char *relpath;
+			dependency_target_specified = true;
+			if (strlen(argv[i]) == 3) {
+				/* -MQ arg or -MT arg */
+				if (i >= argc - 1) {
+					cc_log("Missing argument to %s", argv[i]);
+					stats_update(STATS_ARGS);
+					result = false;
+					goto out;
+				}
+				args_add(dep_args, argv[i]);
+				relpath = make_relative_path(x_strdup(argv[i + 1]));
+				args_add(dep_args, relpath);
+				free(relpath);
+				i++;
+			} else {
+				char *arg_opt;
+				char *option;
+				arg_opt = x_strndup(argv[i], 3);
+				relpath = make_relative_path(x_strdup(argv[i] + 3));
+				option = format("%s%s", arg_opt, relpath);
+				args_add(dep_args, option);
+				free(arg_opt);
+				free(relpath);
+				free(option);
+			}
+			continue;
+		}
+		if (str_startswith(argv[i], "--sysroot=")) {
+			char *relpath = make_relative_path(x_strdup(argv[i] + 10));
+			char *option = format("--sysroot=%s", relpath);
+			args_add(stripped_args, option);
+			free(relpath);
+			free(option);
+			continue;
 		}
 		if (str_startswith(argv[i], "-Wp,")) {
 			if (str_startswith(argv[i], "-Wp,-MD,") && !strchr(argv[i] + 8, ',')) {
@@ -1558,17 +1682,21 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 	output_is_precompiled_header =
 		actual_language && strstr(actual_language, "-header") != NULL;
 
-	if (!found_c_opt && !output_is_precompiled_header) {
-		cc_log("No -c option found");
-		/* I find that having a separate statistic for autoconf tests is useful,
-		   as they are the dominant form of "called for link" in many cases */
-		if (strstr(input_file, "conftest.")) {
-			stats_update(STATS_CONFTEST);
+	if (!found_c_opt) {
+		if (output_is_precompiled_header) {
+			args_add(stripped_args, "-c");
 		} else {
-			stats_update(STATS_LINK);
+			cc_log("No -c option found");
+			/* I find that having a separate statistic for autoconf tests is useful,
+			   as they are the dominant form of "called for link" in many cases */
+			if (strstr(input_file, "conftest.")) {
+				stats_update(STATS_CONFTEST);
+			} else {
+				stats_update(STATS_LINK);
+			}
+			result = false;
+			goto out;
 		}
-		result = false;
-		goto out;
 	}
 
 	if (!actual_language) {
@@ -1667,7 +1795,7 @@ cc_process_args(struct args *orig_args, struct args **preprocessor_args,
 		}
 
 		if (!dependency_target_specified) {
-			args_add(dep_args, "-MT");
+			args_add(dep_args, "-MQ");
 			args_add(dep_args, output_obj);
 		}
 	}
@@ -1710,7 +1838,7 @@ cc_reset(void)
 	base_dir = NULL;
 	args_free(orig_args); orig_args = NULL;
 	free(input_file); input_file = NULL;
-	output_obj = NULL;
+	free(output_obj); output_obj = NULL;
 	free(output_dep); output_dep = NULL;
 	free(cached_obj_hash); cached_obj_hash = NULL;
 	free(cached_obj); cached_obj = NULL;
@@ -1766,6 +1894,29 @@ parse_sloppiness(char *p)
 	return result;
 }
 
+/* Make a copy of stderr that will not be cached, so things like
+   distcc can send networking errors to it. */
+static void
+setup_uncached_err(void)
+{
+	char *buf;
+	int uncached_fd;
+
+	uncached_fd = dup(2);
+	if (uncached_fd == -1) {
+		cc_log("dup(2) failed: %s", strerror(errno));
+		failed();
+	}
+
+	/* leak a pointer to the environment */
+	buf = format("UNCACHED_ERR_FD=%d", uncached_fd);
+
+	if (putenv(buf) == -1) {
+		cc_log("putenv failed: %s", strerror(errno));
+		failed();
+	}
+}
+
 /* the main ccache driver function */
 static void
 ccache(int argc, char *argv[])
@@ -1785,14 +1936,24 @@ ccache(int argc, char *argv[])
 	struct args *compiler_args;
 
 	find_compiler(argc, argv);
+	setup_uncached_err();
 
 	if (getenv("CCACHE_DISABLE")) {
 		cc_log("ccache is disabled");
 		failed();
 	}
 
+	if (!getenv("CCACHE_READONLY")) {
+		if (create_cachedirtag(cache_dir) != 0) {
+			cc_log("failed to create %s/CACHEDIR.TAG (%s)\n",
+			        cache_dir, strerror(errno));
+			failed();
+		}
+	}
+
 	sloppiness = parse_sloppiness(getenv("CCACHE_SLOPPINESS"));
 
+	cc_log_argv("Command line: ", argv);
 	cc_log("Hostname: %s", get_hostname());
 	cc_log("Working directory: %s", current_working_dir);
 
@@ -1801,7 +1962,7 @@ ccache(int argc, char *argv[])
 	}
 
 	if (getenv("CCACHE_UNIFY")) {
-		cc_log("Unify mode disabled");
+		cc_log("Unify mode enabled");
 		enable_unify = true;
 	}
 
@@ -2030,30 +2191,6 @@ ccache_main_options(int argc, char *argv[])
 	return 0;
 }
 
-
-/* Make a copy of stderr that will not be cached, so things like
-   distcc can send networking errors to it. */
-static void
-setup_uncached_err(void)
-{
-	char *buf;
-	int uncached_fd;
-
-	uncached_fd = dup(2);
-	if (uncached_fd == -1) {
-		cc_log("dup(2) failed");
-		failed();
-	}
-
-	/* leak a pointer to the environment */
-	buf = format("UNCACHED_ERR_FD=%d", uncached_fd);
-
-	if (putenv(buf) == -1) {
-		cc_log("putenv failed");
-		failed();
-	}
-}
-
 int
 ccache_main(int argc, char *argv[])
 {
@@ -2079,7 +2216,6 @@ ccache_main(int argc, char *argv[])
 		}
 	}
 
-	current_working_dir = get_cwd();
 	cache_dir = getenv("CCACHE_DIR");
 	if (cache_dir) {
 		cache_dir = x_strdup(cache_dir);
@@ -2120,8 +2256,6 @@ ccache_main(int argc, char *argv[])
 
 	compile_preprocessed_source_code = !getenv("CCACHE_CPP2");
 
-	setup_uncached_err();
-
 	/* make sure the cache dir exists */
 	if (create_dir(cache_dir) != 0) {
 		fprintf(stderr,
@@ -2138,15 +2272,6 @@ ccache_main(int argc, char *argv[])
 		exit(1);
 	}
 
-	if (!getenv("CCACHE_READONLY")) {
-		if (create_cachedirtag(cache_dir) != 0) {
-			fprintf(stderr,
-			        "ccache: failed to create %s/CACHEDIR.TAG (%s)\n",
-			        cache_dir, strerror(errno));
-			exit(1);
-		}
-	}
-
 	ccache(argc, argv);
 	return 1;
 }
