@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 2002 Andrew Tridgell
- * Copyright (C) 2009-2010 Joel Rosdahl
+ * Copyright (C) 2009-2012 Joel Rosdahl
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License as published by the Free
@@ -867,7 +867,8 @@ gnu_getcwd(void)
 		}
 		free(buffer);
 		if (errno != ERANGE) {
-			return 0;
+			cc_log("getcwd error: %d (%s)", errno, strerror(errno));
+			return NULL;
 		}
 		size *= 2;
 	}
@@ -921,6 +922,9 @@ get_cwd(void)
 	struct stat st_cwd;
 
 	cwd = gnu_getcwd();
+	if (!cwd) {
+		return NULL;
+	}
 	pwd = getenv("PWD");
 	if (!pwd) {
 		return cwd;
@@ -932,6 +936,7 @@ get_cwd(void)
 		return cwd;
 	}
 	if (st_pwd.st_dev == st_cwd.st_dev && st_pwd.st_ino == st_cwd.st_ino) {
+		free(cwd);
 		return x_strdup(pwd);
 	} else {
 		return cwd;
@@ -959,7 +964,7 @@ same_executable_name(const char *s1, const char *s2)
 
 /*
  * Compute the length of the longest directory path that is common to two
- * strings.
+ * paths. s1 is assumed to be the path to a directory.
  */
 size_t
 common_dir_prefix_length(const char *s1, const char *s2)
@@ -971,7 +976,21 @@ common_dir_prefix_length(const char *s1, const char *s2)
 		++p1;
 		++p2;
 	}
-	while (p1 > s1 && ((*p1 && *p1 != '/' ) || (*p2 && *p2 != '/'))) {
+	if (*p2 == '/') {
+		/* s2 starts with "s1/". */
+		return p1 - s1;
+	}
+	if (!*p2) {
+		/* s2 is equal to s1. */
+		if (p2 == s2 + 1) {
+			/* Special case for s1 and s2 both being "/". */
+			return 0;
+		} else {
+			return p1 - s1;
+		}
+	}
+	/* Compute the common directory prefix */
+	while (p1 > s1 && *p1 != '/') {
 		p1--;
 		p2--;
 	}
@@ -979,7 +998,9 @@ common_dir_prefix_length(const char *s1, const char *s2)
 }
 
 /*
- * Compute a relative path from from to to. Caller frees.
+ * Compute a relative path from from (an absolute path to a directory) to to (a
+ * path). Assumes that both from and to are well-formed and canonical. Caller
+ * frees.
  */
 char *
 get_relative_path(const char *from, const char *to)
@@ -989,23 +1010,24 @@ get_relative_path(const char *from, const char *to)
 	const char *p;
 	char *result;
 
+	assert(from && from[0] == '/');
+	assert(to);
+
 	if (!*to || *to != '/') {
 		return x_strdup(to);
 	}
 
 	result = x_strdup("");
 	common_prefix_len = common_dir_prefix_length(from, to);
-	for (p = from + common_prefix_len; *p; p++) {
-		if (*p == '/') {
-			x_asprintf2(&result, "../%s", result);
+	if (common_prefix_len > 0 || !str_eq(from, "/")) {
+		for (p = from + common_prefix_len; *p; p++) {
+			if (*p == '/') {
+				x_asprintf2(&result, "../%s", result);
+			}
 		}
 	}
 	if (strlen(to) > common_prefix_len) {
-		p = to + common_prefix_len + 1;
-		while (*p == '/') {
-			p++;
-		}
-		x_asprintf2(&result, "%s%s", result, p);
+		x_asprintf2(&result, "%s%s", result, to + common_prefix_len + 1);
 	}
 	i = strlen(result) - 1;
 	while (i >= 0 && result[i] == '/') {
@@ -1096,15 +1118,19 @@ x_unlink(const char *path)
 	 * file. We don't care if the temp file is trashed, so it's always safe to
 	 * unlink it first.
 	 */
-	const char* tmp_name = format("%s.%s.rmXXXXXX", path, tmp_string());
+	char* tmp_name = format("%s.%s.rmXXXXXX", path, tmp_string());
+	int result = 0;
 	cc_log("Unlink %s via %s", path, tmp_name);
 	if (x_rename(path, tmp_name) == -1) {
-		return -1;
+		result = -1;
+		goto out;
 	}
 	if (unlink(tmp_name) == -1) {
-		return -1;
+		result = -1;
 	}
-	return 0;
+out:
+	free(tmp_name);
+	return result;
 }
 
 #ifndef _WIN32
